--[[
Import libraries, other files.
Save, load, clean and convert data.
Main window with shared elements.
Plugin's command.

DToX, 2013-2014
Licensed under the same terms as Lua itself.
--]]





-- import "Turbine";
-- import "Turbine.UI";
import "Turbine.UI.Lotro";
import "Turbine.Gameplay";

import "DToX.lib.lib";
local Lib = DToX.lib.Lib;
Lib.pName = "Tulkas-M"; -- For wl in lib file
fonts, colors, align = Lib.fonts, Lib.colors, Lib.align;
wl, getBorders, getPosition =
Lib.wl, Lib.getBorders, Lib.getPosition;
-- Plugin
Plugin = Lib.Plugin; -- Functions: language/lang, convertData
-- Tables
Table = Lib.Table; -- Functions: copy, dump, dupes, getKeys, length/len
dt = Table.dump;
-- Visual ones
addBorder = Lib.addBorder;
Hint = Lib.Hint; -- Functions: show, remove
-- Test ones
dv, mu = Lib.dv, Lib.memoryUsage;


-- Variables and constants
--
local pCommandName = "tm";
local pDataName = "TulkasM";
local pDataNameBackup = "TulkasM_Backup";
local pDataNameConverted = "TulkasM_Converted";
local pDataScope = Turbine.DataScope.Account;
data = { -- Contains all data
   ["Items"] = {};
   ["Sets"] = {};
   ["OutdatedNames"] = {};
   ["NotFound"] = {};
};
clickedCheckSets = false;
-- Main's window, carrier and offsets size
local MAIN_CARRIER_WIDTH = 450;
local MAIN_CARRIER_HEIGHT = 50;
local OFFSET = 25;
local OFFSET_TOP = 40;
local MAIN_WINDOW_WIDTH = MAIN_CARRIER_WIDTH + OFFSET * 2;
local MAIN_WINDOW_HEIGHT = MAIN_CARRIER_HEIGHT + OFFSET + OFFSET_TOP;


-- Forward declarations
--
local btn_items;
local btn_sets;
local mainCarrierItems;
local mainCarrierSets;
local windowVeil;
-- Returned from MItems.lua
local btn_updateNames;
local btn_removeItems;
local txt_result;
-- Returned from MSets.lua
local btn_checkSets;





-- Functions
--
-- Tries to load Tulkas DB. Returns true or false. Shows messages
local function loadDB()
   wl( ">> Outsourcing to Russia... Loading DB..." );

   -- From Tulkas (Eng DB; full). Array's name: _ITEMSDB
   local ok = pcall( import, "DToX.bases.Items" );
   -- local ok = pcall( import, "DToX.TulkasToolkit.Items" );

   _ITEMSDB = _ITEMSDB or DToX.bases._ITEMSDB;
   if ( ok ) then
      if ( not _ITEMSDB ) then
         wl( ">> Tulkas' '_ITEMSDB' variable is not found.", "Error" );
         return false;
      else
         local next = next;
         if ( not next( _ITEMSDB ) ) then
            wl( ">> Tulkas' '_ITEMSDB' variable is empty.", "Error" );
            return false;
         else
            wl( ">> DB is ready for torture... Items: " .. Table.len( _ITEMSDB )  .. "." );
            return true;
         end
      end
   else
      wl( ">> No Tulkas' 'Items.lua'. Please, copy it to 'bases' folder.", "Error" );
      return false;
   end
end

-- Shows corresponding tab with visual elements, hides other
local function changeTab( name )
   btn_items:SetEnabled( false );
   btn_sets:SetEnabled( false );
   mainCarrierItems:SetVisible( false );
   mainCarrierSets:SetVisible( false );

   if ( name == "Items" ) then
      btn_sets:SetEnabled( true );
      mainCarrierItems:SetVisible( true );
   elseif ( name == "Sets" ) then
      btn_items:SetEnabled( true );
      mainCarrierSets:SetVisible( true );
   end
end

-- Loads additional data from Tulkas DB.
-- Converts data to Tulkas format and saves it to file
local function convertData()
   -- Check, if Tulkas DB variable is available. If not, try to load
   if ( not _ITEMSDB ) then
      local ok = pcall( import, "DToX.bases.Items" );
      if ( not ok ) then
         wl( ">> No Tulkas' 'Items.lua'. Can't convert.", "Error" );
         return;
      end
      _ITEMSDB = _ITEMSDB or DToX.bases._ITEMSDB;
   end

   local items = data.Items;
   local total = 0;
   local database = {}; -- All items

   -- Get additional items' data from Tulkas DB
   for ID in pairs( items ) do
      local numID = tonumber( ID );
      for k, v in pairs( _ITEMSDB[numID] ) do
         if ( k ~= "Name" and k ~= "Set" ) then
            items[ID][k] = v;
         end
      end
   end

   -- Convert items' data to strings
   for ID, v in pairs( items ) do
      local line = ""; -- For one item

      -- Correct name. Replace "’" with "'"
      if ( v.Name:find( "s’ " ) ) then
         v.Name = v.Name:gsub( "s’ ", "s' " );
      end
      if ( v.Name:find( "’s" ) ) then
         v.Name = v.Name:gsub( "’s", "'s" );
      end
      if ( v.Name:find( "�s" ) ) then
         v.Name = v.Name:gsub( "�s", "'s" );
      end

      -- Get stats part
      local stats = "";
      for statName, statValue in pairs( v.Stats ) do
         if ( statName == "Damage" ) then
            local dmg = v.Stats.Damage;
            -- Tulkas uses for in loop, so order will be off
            stats = stats
            .. '["Damage"]={'
               .. '["TypeDMG"]="' .. dmg.TypeDMG .. '";'
               .. '["MinDMG"]=' .. dmg.MinDMG .. ';'
               .. '["MaxDMG"]=' .. dmg.MaxDMG .. ';'
               .. '["DPS"]=' .. dmg.DPS .. ';'
            .. '};'
         else
            stats = stats .. '["' .. statName .. '"]=' .. statValue .. ';'
         end
      end
      if ( stats:find( "," ) ) then
         stats = stats:gsub( ",", "." );
      end

      -- Get set part
      local set = "";
      if ( v.Set ) then
         -- Sort bonuses' number of equipped items
         local sortedNum = {};
         for i in pairs( v.Set.Bonus ) do
            local j = #sortedNum + 1;
            sortedNum[j] = i;
         end
         table.sort( sortedNum );

         -- Set: get Bonus part
         local bonus = "";
         for _, i in pairs( sortedNum ) do
            -- Set: bonuses for N equipped items
            local iBonus = "";
            for bonusKey, bonusValue in pairs( v.Set.Bonus[i] ) do
               iBonus = iBonus .. '["' .. bonusKey .. '"]="' .. bonusValue .. '";'
            end

            -- Set: bonuses for N equipped items (full line)
            bonus = bonus .. '[' .. i .. '] = {' .. iBonus .. '};';
         end

         -- Set: full line
         set = 
         '["Set"] = {'
            .. '["Name"]="' .. v.Set.Name .. '";'
            .. '["IDs"] = {' .. table.concat( v.Set.IDs, "," ) .. '};'
            .. '["Bonus"] = {' .. bonus .. '};'
         .. '};';
      end

      -- Get full line
      line =
      '[' .. ID .. '] = {'
         .. '["Name"]="' .. v.Name .. '";'
         .. '["Level"]=' .. v.Level .. ';'
         .. '["ItemLevel"]=' .. v.ItemLevel .. ';'
         .. '["Slot"]="' .. v.Slot .. '";'
         .. '["SubSlot"]="' .. v.SubSlot .. '";'
         .. '["Type"]="' .. v.Type .. '";'
         .. '["Class"]="' .. v.Class .. '";'
         .. '["Color"]="' .. v.Color .. '";'
         .. '["Stats"] = {' .. stats .. '};'
         .. set
      .. '};';

      -- Prepare for saving
      if ( line:find( "'" ) ) then
         line = line:gsub( "'", "__" );
      end
      line = line:gsub( '"', "#" );

      database[#database + 1] = line;
      total = total + 1;
   end

   saveData( pDataNameConverted, database );
   wl( ">> Converted items: " .. total .. "."
   .. "\nSaved to file ...\\<Documents>\\The Lord of the Rings Online\\PluginData\\<YourLogin>\\AllServers\\" .. pDataNameConverted .. ".plugindata."
   .. "\nUse supplied *.html file to clean it and get Tulkas DB." );
end

-- Removes raw sets, corrects indexes, returns result
local function cleanData()
   local cData = Table.copy( data );
   -- Remove raw sets, correct indexes
   local tmp = {};
   for i, v in pairs( cData.Sets ) do
      if ( not v.isRaw ) then
         local j = #tmp + 1;
         tmp[j] = v;
      end
   end
   cData.Sets = tmp;

   -- Remove all fields except Name and Set
   for ID, info in pairs( cData.Items ) do
      for k, v in pairs( info ) do
         if ( k ~= "Name" and k ~= "Set"  ) then
            cData.Items[ID][k] = nil;
         end
      end
   end

   cData = Plugin.convertData( cData, "To" );

   return cData;
end

local function loadBackup( savedData )
   if ( savedData ) then
      wl( ">> Loaded backup." );
      data = Plugin.convertData( savedData, "From" );
   else
      wl( ">> Backup is empty ¯\\(º_o)/¯", "Error" );
   end

   -- Update lists with sets and hide veil
   if ( clickedCheckSets ) then
      btn_checkSets.MouseClick();
   end
   windowVeil:SetVisible( false );
   windowVeil:SetParent( nil );
end

-- Add shortcut to quickslot, return it
function addShortcut( quickslot, decID )
   local hex = string.format( "%X", decID );
   local sc = Turbine.UI.Lotro.Shortcut( Turbine.UI.Lotro.ShortcutType.Item, "0x0,0x" .. hex );
   quickslot:SetShortcut( sc );
   return sc;
end

function saveData( name, data )
   Turbine.PluginData.Save( pDataScope, name, data );
end

function loadData( name, onLoad )
   return Turbine.PluginData.Load( pDataScope, name, onLoad );
end

-- Clears sets and saves backup
function Turbine.Plugin.Unload()
   local data = cleanData();
   saveData( pDataNameBackup, data );
   wl( ">> Saved backup." );
end




-- Visuals
--

-- Main window
win_main = Turbine.UI.Lotro.Window();
win_main:SetText( "Madness of... Tulkas" );
win_main:SetVisible( true );
win_main:SetSize( MAIN_WINDOW_WIDTH, MAIN_WINDOW_HEIGHT );
local left = 200;
local top = 100;
win_main:SetPosition( left, top );
function win_main:Closed()
   wl( "To show window use command: /tm or /tm show." );
end
-- Resize veil
function win_main:SizeChanged()
   local w, h = win_main:GetSize();
   windowVeil:SetSize( w, h );
end

-- Main carrier for elements (offseted from window's borders)
local mainCarrier = Turbine.UI.Control();
mainCarrier:SetParent( win_main );
mainCarrier:SetSize( MAIN_CARRIER_WIDTH, MAIN_CARRIER_HEIGHT );
mainCarrier:SetPosition( OFFSET, OFFSET_TOP );

-- Main carriers for elements from other files
local b = getBorders( mainCarrier );
local left = b.Left;
local top = b.Bottom + 20;
mainCarrierItems = Turbine.UI.Control();
mainCarrierItems:SetParent( win_main );
mainCarrierItems:SetPosition( left, top );
mainCarrierSets = Turbine.UI.Control();
mainCarrierSets:SetParent( win_main );
mainCarrierSets:SetPosition( left, top );

-- Button: Work with items
btn_items = Turbine.UI.Lotro.Button();
btn_items:SetParent( mainCarrier );
btn_items:SetFont( fonts.Verdana14 );
btn_items:SetText( "Work with items" );
btn_items:SetSize( 130, 20 );
btn_items:SetPosition( 0, 0 );
-- Show tab, load file, return elements for manipulation
function btn_items:MouseClick()
   changeTab( "Items" );
   import "DToX.TulkasToolkit.MItems";
   txt_result, btn_updateNames, btn_removeItems = showItems( mainCarrierItems );
end

-- Button: Work with sets
btn_sets = Turbine.UI.Lotro.Button();
btn_sets:SetParent( mainCarrier );
btn_sets:SetFont( fonts.Verdana14 );
btn_sets:SetText( "Work with sets" );
btn_sets:SetSize( 130, 20 );
local b = getBorders( btn_items );
btn_sets:SetPosition( b.Right + 30, b.Top );
-- Show tab, load file, return elements for manipulation
function btn_sets:MouseClick()
   changeTab( "Sets" );
   import "DToX.TulkasToolkit.MSets";
   btn_checkSets = showSets( mainCarrierSets );
end

-- Button: Convert data
local btn_convertData = Turbine.UI.Lotro.Button();
btn_convertData:SetParent( mainCarrier );
btn_convertData:SetFont( fonts.Verdana14 );
btn_convertData:SetText( "Convert data" );
btn_convertData:SetSize( 130, 20 );
local b = getBorders( btn_sets );
btn_convertData:SetPosition( b.Right + 30, b.Top );
function btn_convertData:MouseHover()
   Hint.show( self, "Convert current data to Tulkas format. Result'll be saved in file. Then use supplied *.html file to clean it and get Tulkas DB.");
end
-- Check sets, load additional data from Tulkas DB, convert
function btn_convertData:MouseClick()
   -- Check sets
   if ( not btn_checkSets ) then
      btn_sets.MouseClick();
      btn_items.MouseClick();
      btn_checkSets.MouseClick();
   end

   convertData();
end

-- Button: Load DB
local btn_loadDB = Turbine.UI.Lotro.Button();
btn_loadDB:SetParent( mainCarrier );
btn_loadDB:SetFont( fonts.Verdana14 );
btn_loadDB:SetText( "Load database" );
btn_loadDB:SetSize( 130, 20 );
local b = getBorders( btn_items );
local left = b.Right - btn_items:GetWidth() / 2;
btn_loadDB:SetPosition( left, b.Bottom + 10 );
function btn_loadDB:MouseHover()
   Hint.show( self, "Load data from Tulkas Items.lua file.\n\nWarning! It'll rewrite current items." );
end
-- Load DB or show message, prepare elements on Items tab, re-check sets
function btn_loadDB:MouseClick()
   if ( loadDB() ) then
      data.Items = Table.copy( _ITEMSDB );
      btn_loadDB:SetEnabled( false );
   end

   if ( btn_updateNames ) then
      btn_updateNames:SetEnabled( false );
      btn_removeItems:SetEnabled( false );
      txt_result:SetText( "" );
   end

   if ( clickedCheckSets ) then
      btn_checkSets.MouseClick();
   end
end

-- Button: Save data
local btn_saveData = Turbine.UI.Lotro.Button();
btn_saveData:SetParent( mainCarrier );
btn_saveData:SetFont( fonts.Verdana14 );
btn_saveData:SetText( "Save data" );
btn_saveData:SetSize( 130, 20 );
local b = getBorders( btn_sets );
local left = b.Right - btn_sets:GetWidth() / 2;
btn_saveData:SetPosition( left, b.Bottom + 10 );
-- Clean sets, save data, show path to file
function btn_saveData:MouseClick()
   local data = cleanData();
   saveData( pDataName, data ); -- !! Improve time?
   wl( ">> Saved to file ...\\<Documents>\\The Lord of the Rings Online\\PluginData\\<YourLogin>\\AllServers\\" .. pDataName .. ".plugindata." );
end

-- Veil: lies over window
windowVeil = Turbine.UI.Label();
windowVeil:SetParent( win_main );
windowVeil:SetBackColor( colors.Grey05 );
windowVeil:SetBackColorBlendMode( 3 );
windowVeil:SetTextAlignment( align[5] );
windowVeil:SetFont( fonts.Verdana23 );
windowVeil:SetText( "Loading..." );
windowVeil:SetVisible( false );





-- Add command
--
pCommand = Turbine.ShellCommand();
Turbine.Shell.AddCommand( pCommandName, pCommand );
function pCommand:Execute( cmd, args )
   if ( args == "" or args == "show" ) then
      win_main:SetVisible( true );
      win_main:Activate();

   elseif ( args == "?" ) then
      local c = "/" .. cmd;
      wl ( ">> Available commands:"
      .. "\n" .. c .. " ? - shows this message."
      .. "\n" .. c .. " or " .. c .. " show - shows window."
      .. "\n" .. c .. " load backup - loads backup, which is saved when you unload plugin."
      .. "\n" .. c .. " clear data - removes all items and sets. If something went wrong and you don't want to wait for backup to be saved. Or just don't want to wait and all data was saved manually."
      .. "\n" .. c .. " clear items - removes all items."
      .. "\n" .. c .. " clear sets - removes all sets."
      .. "" );

   elseif ( args == "load backup" ) then
      loadData( pDataNameBackup, loadBackup );
      wl( ">> Loading backup... please, wait ~15 seconds." );
      windowVeil:SetParent( win_main );
      windowVeil:SetVisible( true );

   elseif ( args == "clear data" ) then
      data.Items = {};
      data.Sets = {};
      wl( ">> All items and sets are removed. To reverse it, unload plugin. Don't press 'Save data'.", "Warning" );
      if ( clickedCheckSets ) then
         btn_checkSets.MouseClick();
      end

   elseif ( args == "clear sets" ) then
      data.Sets = {};
      wl( ">> All saved sets are removed. To reverse it, unload plugin. Don't press 'Save data'.", "Warning" );
      if ( clickedCheckSets ) then
         btn_checkSets.MouseClick();
      end

   elseif ( args == "clear items" ) then
      data.Items = {};
      wl( ">> All items are removed." );

   -- Moves sets, that have filled Set.Bonus key to the edited list
   -- For database from M-13 Project
   elseif ( args == "mcs" ) then
      -- Load tab and create set's list
      if ( not clickedCheckSets ) then
         btn_sets.MouseClick();
         btn_checkSets.MouseClick();
      end

      local sets = data.Sets;
      local items = data.Items;
      for setIndex, setData in pairs( sets ) do
         local numID = setData.ID;
         local item = items[numID];

         if ( item.Set and item.Set.Bonus ) then
            local setBonusesNum = Table.len( item.Set.Bonus );
            if ( setBonusesNum > 0 ) then
               sets[setIndex].Set = item.Set;
               sets[setIndex].isRaw = false;
            end
         end
      end

      -- Update sets' lists
      if ( clickedCheckSets ) then
         btn_checkSets.MouseClick();
      end

      wl( ">> Sets with bonuses moved to edited list." );

   else
      wl( "Error. Please, check command.", "Error" );
   end
end





-- Logic
--
-- Open items tab
btn_items.MouseClick();
-- btn_sets.MouseClick();

-- Load saved data
local savedData = loadData( pDataName );
if ( savedData ) then
   wl( ">> Loaded saved data." );
   data = Plugin.convertData( savedData, "From" );
else
   wl( ">> No saved data." );
   btn_loadDB.MouseClick();
end

wl( "\nWelcome. Don't forget to press 'Save data', when you're done."
.. "\nTo remove all saved sets use command: /tm clear sets."
.. "\nTo load backup use command: /tm load backup. Backup is saved, when you unload the plugin."
.. "\nTo show available commands use command: /tm ?."
);